local COLOR_BORDER = {0.5, 0.5, 0.5};
local COLOR_INDICATOR = {1, 0.5, 0};
local COLOR_SHINE = ORANGE_FONT_COLOR;

local THRESHOLD = 0.01;
local SCROLLTIME = 0.3;
local DEFAULT_TEMPLATE = "MessIconElementTemplate";
local ELEMENT_WIDTH = 32;
local ELEMENT_HEIGHT = 32;
local ELEMENT_GAP = 2;
local DEFAULT_GRIDSIZE = 5;
local SHINESCALE = 1.5;

local Selector_OnDeselected = "Selector_OnDeselected";
local Selector_OnSelected = "Selector_OnSelected";

local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local function mathRound(num)
	return mathFloor(num + .5);
end

local function mathBound(value, min, max)
	return mathMax(min, mathMin(max, value));
end

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

----------------------------------------------------------------	QueueTemplate
--	non-duplicated FIFO queue, with default capacity 1. (negative capacity for unlimited queue)
local QueueTemplate = {
	["_size"] = 0,
	["_capacity"] = 1,
	["_itemList"] = nil,		--	index	:	item
	["_itemMap"] = nil,			--	item	:	isExisted
};

function QueueTemplate:_overflow()
	if self._capacity < 0 then
		--	unlimited
	elseif self._size > self._capacity then
		local overflowed = {};
		while self._size > self._capacity do
			local item = tremove(self._itemList, 1);
			self._itemMap[item] = nil;
			self._size = self._size - 1;
			tinsert(overflowed, item);
		end
		return overflowed;
	end
end

function QueueTemplate:New(queue)
	queue = setmetatable(queue or {}, {__index = self});
	queue._itemList = {};
	queue._itemMap = {};
	return queue;
end

function QueueTemplate:SetMaxSize(size)		--	returns out of capacity items (in FIFO order).
	self._capacity = size;
	return self:_overflow();
end

function QueueTemplate:GetMaxSize()
	return self._capacity;
end

function QueueTemplate:GetSize()
	return self._size;
end

function QueueTemplate:Exists(item)
	return self._itemMap[item];
end

function QueueTemplate:Get(index)			--	get specific index item or the first one.
	return self._itemList[index or 1];
end

function QueueTemplate:Add(item)
	if self:Exists(item) then return; end
	tinsert(self._itemList, item);
	self._itemMap[item] = true;
	self._size = self._size + 1;
	return self:_overflow();
end

function QueueTemplate:Remove(item)			--	remove specific item or the first one.
	if not item then
		item = tremove(self._itemList, 1);
		self._itemMap[item] = nil;
		self._size = self._size - 1;
		return item;
	elseif self:Exists(item) then
		for index, it in ipairs(self._itemList) do
			if it == item then
				tremove(self._itemList, index);
				self._itemMap[item] = nil;
				self._size = self._size - 1;
				return item;
			end
		end
	end
end

function QueueTemplate:Clear()				--	empty queue.
	wipe(self._itemList);
	wipe(self._itemMap);
	self._size = 0;
end

----------------------------------------------------------------	AutoCastShineTemplate
local castShinePool = {};
local shineCount = 0;

local function newShine(parent)
	local s = tremove(castShinePool);
	if s then
		s:SetParent(parent);
	else
		shineCount = shineCount + 1;
		s = CreateFrame("Frame", "MessShine"..shineCount, parent, "AutoCastShineTemplate");
		s:SetFrameStrata("TOOLTIP");
		for _, sparks in pairs(s.sparkles) do
			sparks:SetSize(sparks:GetWidth() * SHINESCALE, sparks:GetHeight() * SHINESCALE);
		end
	end
	return s;
end

local function returnShine(s)
	AutoCastShine_AutoCastStop(s);
	s:Hide();
	s:ClearAllPoints();
	tinsert(castShinePool, s);
end

local function applyShine(target, shine, colorShine)
	shine = shine or newShine(target);
	colorShine = colorShine or COLOR_SHINE;
	shine:SetParent(target);
	shine:SetAllPoints(target);
	shine:Show();
	AutoCastShine_AutoCastStart(shine, unpack(colorShine));
	return shine;
end

----------------------------------------------------------------	MessIconElementTemplate
function MessIconElement_OnLoad(self)
	self:RegisterForClicks("AnyDown");
	self.highlight:SetVertexColor(unpack(COLOR_INDICATOR));
end

function MessIconElement_OnEnter(self, motion)
	self.highlightAnime:Play();
end

function MessIconElement_OnLeave(self, motion)
	self.highlightAnime:Stop();
end

----------------------------------------------------------------	MassData
local MassData = {
	["name"] = "",
	["size"] = 0,
	["list"] = nil,	--	index : data
	["map"] = nil,	--	data : index
}

function MassData:New(datas)
	datas = datas or {};
	setmetatable(datas, {__index = self});
	datas.list = {};
	datas.map = {};
	return datas;
end

function MassData:Consume(name, list)	--	also remove duplicated data
	self.name = name or "";
	local size = 0;
	for index, data in ipairs(list) do
		if not self.map[data] then
			tinsert(self.list, data);
			size = size + 1;
			self.map[data] = size;
		end
	end
	self.size = size;
end

function MassData:Get(index)
	return self.list[index];
end

function MassData:IndexOf(data)
	return self.map[data];
end

function MassData:First()
	return self:Get(1);
end

function MassData:Next(data)
	local index = self:IndexOf(data);
	if index and index < self.size then
		return self:Get(index + 1);
	end
end

----------------------------------------------------------------	MessGridcMassSelectorTemplate
local MessGridMassSelectorMeta = {
	--	caches
	["_row"] = nil,
	["_col"] = nil,
	--	attributes
	["_dataAltered"] = nil,
	["_pool"] = nil,			--	index : element
	["_selections"] = nil,		--	QueueTemplate
	["_shines"] = nil,			--	data : shine
	["_elements"] = nil,		--	index : element
	["_elementDataMap"] = nil,	--	element : data
	["_massList"] = nil,		--	index : massdata
	["_template"] = DEFAULT_TEMPLATE,
	["_deselectable"] = nil,
	["elementWidth"] = ELEMENT_WIDTH,
	["elementHeight"] = ELEMENT_HEIGHT,
	["elementGap"] = ELEMENT_GAP,
	["selectedColor"] = COLOR_SHINE,
	["scrollStep"] = 1,
	--	private
	["_newElement"] = function(self, parent)
		local e = tremove(self._pool);
		if e then
			e:SetParent(parent);
		else
			e = CreateFrame("Button", nil, parent, self._template);
		end
		return e;
	end,
	["_elementOnClick"] = function(self, button)
		local parnet = self:GetParent();
		local data = parnet._elementDataMap[self];
		if parnet:HasSelection(data) then
			if parnet:IsDeselectable() then
				parnet:RemoveSelection(data, button);
			end
		else
			parnet:AddSelection(data, button);
		end
	end,
	["_updateViewport"] = function(self, row, col)		--	update viewport size & bar minmax
		local w, h = self:GetSize();
		local unitWidth = self.elementWidth + self.elementGap;
		local unitHeight = self.elementHeight + self.elementGap;
		if row then
			h = row * unitHeight + self.elementGap;
		else
			row = mathFloor((h - self.elementGap)/ unitHeight);
		end
		if col then
			w = col * unitWidth + self.elementGap;
		else
			col = mathFloor((w - self.elementGap)/ unitWidth);
		end
		
		self.anchor:SetSize(
			mathMax(1e-9, (w - col * unitWidth + self.elementGap) / 2), 
			mathMax(1e-9, (h - row * unitHeight + self.elementGap) / -2));
		if row ~= self._row or col ~= self._col then
			self:_rebuildElements(row, col);
			self._row = row;
			self._col = col;
		end
		
		local totalRow = 0;
		for i, mass in ipairs(self._massList) do
			totalRow = totalRow + mathCeil(mass.size / col);
		end
		self.bar:SetViewportSize(col, row);
		self.bar:SetMinMaxValues(0, mathMax(0, totalRow - row - 1));
		
		self.scrollStep = 1;
		self._dataAltered = nil;
	end,
	["_rebuildElements"] = function(self, row, col)
		--	clean
		for _, e in ipairs(self._elements) do
			e:Hide();
			e:ClearAllPoints();
			tinsert(self._pool, e);
		end
		wipe(self._elements);
		if not row or not col or row < 1 or col < 1 then return; end
		--	rebuild
		for y = 1, row do
			--	leading
			local e = self:_newElement(self);
			e:SetScript("OnClick", self._elementOnClick);
			e:SetSize(self.elementWidth, self.elementHeight);
			e:SetPoint("TOPLEFT", self.anchor, "BOTTOMRIGHT", 0, (1 - y) * self.elementHeight - y * self.elementGap);
			tinsert(self._elements, e);
			--	siblings
			for x = 2, col do
				local es = self:_newElement(self);
				es:SetScript("OnClick", self._elementOnClick);
				es:SetSize(self.elementWidth, self.elementHeight);
				es:SetPoint("LEFT", e, "LEFT",	(x - 1) * (self.elementWidth + self.elementGap), 0);
				tinsert(self._elements, es);
			end
		end
	end,
	["_getIndexByData"] = function(self, data)		--	first match
		for i, mass in ipairs(self._massList) do
			local index = mass:IndexOf(data);
			if index then
				return i, index;
			end
		end
	end,
	["_getMass"] = function(self, massIndex)
		return self._massList[massIndex];
	end,
	["_scrollTo"] = function(self, massIndex, dataIndex)
		self:Refresh();
		if not massIndex or not dataIndex then return; end

		local col, row = self.bar:GetViewportSize();
		local valueNow = self.bar:GetValue();
		local valueNew = 0;
		for i = 1, massIndex - 1 do
			local mass = self._massList[i];
			valueNew = valueNew + mathCeil(mass.size / col);
		end
		valueNew = valueNew + mathCeil(dataIndex / col) - 1;	--	scrolling value begins at 0
		
		if valueNew < valueNow then
			--	upward
			self.bar:SetValue(valueNew);
			self:Refresh();
		elseif valueNow < (valueNew - row + 1) then
			--	downward
			self.bar:SetValue(valueNew - row + 1);
			self:Refresh();
		else
			--	inbound, do nothing
		end
	end,
	--	public
	["GetSelectLimit"] = function(self)
		return self._selections:GetMaxSize();
	end,
	["SetSelectLimit"] = function(self, num)
		local overflows = self._selections:SetMaxSize(num);
		if overflows then
			for _, overflow in ipairs(overflows) do
				self:_dispatch(Selector_OnDeselected, overflow);
			end
		end
		self:Refresh();
	end,
	["IsDeselectable"] = function(self)
		return self._deselectable;
	end,
	["SetDeselectable"] = function(self, flag)
		self._deselectable = flag;
	end,
	["HasSelection"] = function(self, data)
		return self._selections:Exists(data);
	end,
	["RemoveSelection"] = function(self, data, motion)
		self._selections:Remove(data);
		self:_dispatch(Selector_OnDeselected, data, motion);
		self:Refresh();
	end,
	["AddSelection"] = function(self, data, motion)
		local overflows = self._selections:Add(data);
		if overflows then
			for _, overflow in ipairs(overflows) do
				self:_dispatch(Selector_OnDeselected, overflow);
			end
		end
		self:_dispatch(Selector_OnSelected, data, motion);
		self:Refresh();
	end,
	["GetSelection"] = function(self, index)
		return self._selections:Get(index);
	end,
	["ClearSelection"] = function(self)
		self._selections:Clear();
		for data, shine in pairs(self._shines) do
			returnShine(shine);
		end
		wipe(self._shines);
	end,
	["SetTemplate"] = function(self, name)
		if name == self._template then return; end
		self._template = name or DEFAULT_TEMPLATE;
		for index, e in ipairs(self._elements) do
			e:ClearAllPoints();
			e:Hide();
		end
		wipe(self._elements);
		for index, e in ipairs(self._pool) do
			e:ClearAllPoints();
			e:Hide();
		end
		wipe(self._pool);
		self:_rebuildElements(self._row, self._col);
		self:Refresh();
	end,
	["SetPreferedSize"] = function(self, row, col)
		local w = col * self.elementWidth + (col + 1) * self.elementGap + 0.1;
		local h = row * self.elementHeight + (row + 1) * self.elementGap + 0.1;
		self:SetSize(w, h);
		self:_updateViewport(row, col);
		self:Refresh();
		return w, h;
	end,
	["Refresh"] = function(self)				--	updates elements & manipulates shines
		if self._dataAltered or not self._row or not self._col then
			self:_updateViewport();
		end
		
		local col, row = self.bar:GetViewportSize();
		local rowIndex = mathRound(self.bar:GetValue()) + 1;
		local massIndex, dataIndex;
		for i, mass in ipairs(self._massList) do
			local rowCount = mathCeil(mass.size / col);
			rowIndex = rowIndex - rowCount;
			if rowIndex <= 0 then
				massIndex = i;
				dataIndex = (rowIndex + rowCount - 1) * col + 1;
				break;
			end
		end
		if not massIndex or not dataIndex then return; end
		local mass = self:_getMass(massIndex);
		local data = mass:Get(dataIndex);

		--	update elements
		wipe(self._elementDataMap);
		local skipped, notEmpty = 0;
		for i, e in ipairs(self._elements) do
			if skipped > 0 then			--	skip to newline (higher priority)
				skipped = skipped - 1;
				e:Hide();
			elseif data then			--	there is data, visualize it
				notEmpty = true;
				self:Visualize(e, data);
				self._elementDataMap[e] = data;
				--	selected shine
				if self:HasSelection(data) then
					self._shines[data] = applyShine(e, self._shines[data], self.selectedColor);
					self._shines[data].valid = true;
				end
				--	update mouseover
				pcall(e:IsMouseOver() and e:GetScript("OnEnter") or e:GetScript("OnLeave"), e, false);
				data = mass:Next(data);
				--	if is last data
				if not data and massIndex < #self._massList then
					massIndex = massIndex + 1;
					mass = self:_getMass(massIndex);
					data = mass:First();
					local leftCol = i % col;
					leftCol = leftCol == 0 and leftCol or col - leftCol;
					skipped = skipped + leftCol;
				end
			else
				e:Hide();
			end
		end
		--	recycle invalid shine
		for data, shine in pairs(self._shines) do
			if shine.valid then
				shine.valid = nil;
			else
				self._shines[data] = nil;
				returnShine(shine);
			end
		end
		--	empty indicator
		setVisible(self.emptyText, not notEmpty);
	end,
	["MountMassData"] = function(self, name, list, nonTerminated)
		local mass = MassData:New();
		mass:Consume(name, list);
		tinsert(self._massList, mass);
		self._dataAltered = true;
		if not nonTerminated then
			self:Refresh();
		end
	end,
	["ScrollToData"] = function(self, data)
		self:_scrollTo(self:_getIndexByData(data));
	end,
	["SetScrollValue"] = function(self, value, orintation)
		if not orintation or orintation == "VERTICAL" then
			value = mathRound(value);
			self.bar:SetValue(mathBound(value, self.bar:GetMinMaxValues()));
			self:Refresh();
		end
	end,
	["Visualize"] = function(self, element, data)
		element.icon:SetTexture(data);
	end,
};

function MessGridMassSelector_OnMouseWheel(self, delta)
	local valueNew = self.bar:GetValue() - delta * self.scrollStep;
	valueNew = mathBound(valueNew, self.bar:GetMinMaxValues());
	self.bar:SetValue(mathRound(valueNew));
	self:Refresh();
end

function MessGridMassSelector_OnSizeChanged(self, width, height)
	if not width or not height then return; end
	local row = mathFloor((height - self.elementGap)/ (self.elementHeight + self.elementGap));
	local col = mathFloor((width - self.elementGap)/ (self.elementWidth + self.elementGap));
	self:_updateViewport(row, col);
	self:Refresh();
end

function MessGridMassSelector_OnLoad(self)
	MessEventDispatcher_OnLoad(self);
	inherits(self, MessGridMassSelectorMeta);

	self._pool = {};
	self._selections = QueueTemplate:New();
	self._shines = {};
	self._elements = {};
	self._elementDataMap = {};
	self._massList = {};

	self.LBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.RBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.BBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.TBorder:SetVertexColor(unpack(COLOR_BORDER));

	self.bar:SetScrollParent(self);
	self.bar:SetPreferSmooth(false);
	self.bar:SetFadeAlpha(1, 0.5);
	self.bar:SetBorderColor(COLOR_BORDER);
end

----------------------------------------------------------------	MessFauxMassSelectorTemplate
local MessFauxMassSelectorMeta = {
	--	caches
	["_row"] = nil,
	["_col"] = nil,
	["_elapsed"] = 0,			--	time since last updated
	["_smoothTimer"] = nil,		--	smooth duration left
	["_smoothVelocity"] = 0,	--	smooth velocity (vector)
	--	attributes
	["_dataAltered"] = nil,		--	mounted data altered flag
	["_pool"] = nil,			--	index : element
	["_selections"] = nil,		--	(QueueTemplate) :	index : selected data
	["_shines"] = nil,			--	data : shine
	["_elements"] = nil,		--	index : element
	["_elementDataMap"] = nil,	--	element : data
	["_massList"] = nil,		--	index : massdata
	["_template"] = DEFAULT_TEMPLATE,
	["_deselectable"] = nil,
	["elementWidth"] = ELEMENT_WIDTH,
	["elementHeight"] = ELEMENT_HEIGHT,
	["elementGap"] = ELEMENT_GAP,
	["selectedColor"] = COLOR_SHINE,
	["scrollStep"] = ELEMENT_WIDTH + ELEMENT_GAP,
	--	private
	["_newElement"] = function(self, parent)
		local e = tremove(self._pool);
		if e then
			e:SetParent(parent);
		else
			e = CreateFrame("Button", nil, parent, self._template);
		end
		return e;
	end,
	["_elementOnClick"] = function(self, button)
		local parnet = self:GetParent():GetParent();
		local data = parnet._elementDataMap[self];
		if parnet:HasSelection(data) then
			if parnet:IsDeselectable() then
				parnet:RemoveSelection(data, button);
			end
		else
			parnet:AddSelection(data, button);
		end
	end,
	["_updateViewport"] = function(self, row, col)		--	update viewport size & bar minmax, force row, col if assigned.
		local w, h = self:GetSize();
		local unitWidth = self.elementWidth + self.elementGap;
		local unitHeight = self.elementHeight + self.elementGap;
		if row then
			h = row * unitHeight + self.elementGap;
		else
			row = mathFloor((h - self.elementGap)/ unitHeight);
		end
		if col then
			w = col * unitWidth + self.elementGap;
		else
			col = mathFloor((w - self.elementGap)/ unitWidth);
		end
		
		self.content.anchor:SetSize(w, mathMax(1, (h - row * unitHeight + self.elementGap) / 2));
		if row ~= self._row or col ~= self._col then
			self:_rebuildElements(row, col);
			self._row = row;
			self._col = col;
		end
		
		local colCount = 0;
		for i, mass in ipairs(self._massList) do
			colCount = colCount + mathCeil(mass.size / row);
		end
		self.bar:SetViewportSize(w, h);
		self.bar:SetMinMaxValues(0, mathMax(0, (colCount - col) * unitWidth + self.elementGap));
		
		self.scrollStep = unitWidth;
		self._dataAltered = nil;
	end,
	["_rebuildElements"] = function(self, row, col)
		--	clean
		for _, e in ipairs(self._elements) do
			e:Hide();
			e:ClearAllPoints();
			tinsert(self._pool, e);
		end
		wipe(self._elements);

		if not row or not col or row < 1 or col < 1 then return; end
		--	rebuild
		for x = 1, col + 2 do
			--	leading
			local e = self:_newElement(self.content);
			e:SetScript("OnClick", self._elementOnClick);
			e:SetSize(self.elementWidth, self.elementHeight);
			e:SetPoint("TOPLEFT", self.content.anchor, "BOTTOMLEFT", (x - 1) * self.elementWidth + x * self.elementGap, 0);
			tinsert(self._elements, e);
			--	siblings
			for y = 2, row do
				local es = self:_newElement(self.content);
				es:SetScript("OnClick", self._elementOnClick);
				es:SetSize(self.elementWidth, self.elementHeight);
				es:SetPoint("TOP", e, "TOP", 0, (1 - y) * (self.elementHeight + self.elementGap));
				tinsert(self._elements, es);
			end
		end
	end,
	["_getIndexByOffset"] = function(self, offset)
		local col, row = self._col, self._row;
		local unitWidth = self.elementWidth + self.elementGap;
		
		local colIndex = mathMax(1, mathCeil(offset / unitWidth));
		for i, mass in ipairs(self._massList) do
			local colCount = mathCeil(mass.size / row);
			colIndex = colIndex - colCount;
			if colIndex <= 0 then
				return i, (colIndex + colCount - 1) * row + 1, offset % unitWidth;
			end
		end
		--	not found (error)
	end,
	["_getIndexByData"] = function(self, data)		--	first match
		for i, mass in ipairs(self._massList) do
			local index = mass:IndexOf(data);
			if index then
				return i, index;
			end
		end
	end,
	["_getMass"] = function(self, massIndex)
		return self._massList[massIndex];
	end,
	["_setOffset"] = function(self, offset)
		offset = mathBound(offset, self.bar:GetMinMaxValues());
		self.bar:SetValue(offset);
		
		local massIndex, dataIndex, fauxOffset = self:_getIndexByOffset(offset);
		self:SetHorizontalScroll(fauxOffset);
		--	update faux display, if necessay
		if self._viewMassIndex ~= massIndex or self._viewDataIndex ~= dataIndex then
			self._viewMassIndex = massIndex;
			self._viewDataIndex = dataIndex;
			self:Refresh(massIndex, dataIndex);
		end
	end,
	["_scrollTo"] = function(self, massIndex, dataIndex)
		self:Refresh();
		if not massIndex or not dataIndex then return; end
		
		local col, row = self._col, self._row;
		local w, h = self.bar:GetViewportSize();
		local unitWidth = self.elementWidth + self.elementGap;
		local valueNow = self.bar:GetValue();
		local valueNew = 0;
		for i = 1, massIndex - 1 do
			local mass = self._massList[i];
			valueNew = valueNew + mathCeil(mass.size / row);
		end
		valueNew = valueNew + mathCeil(dataIndex / row) - 1;	--	scrolling value begins at 0
		valueNew = valueNew * unitWidth;
		
		if valueNew < valueNow then
			--	backward
			self:HorizontalSmoothScrollTo(valueNew);
		elseif valueNow < (valueNew + unitWidth - w) then
			--	forward
			self:HorizontalSmoothScrollTo(valueNew + unitWidth - w);
		else
			--	inbound, do nothing
		end
	end,
	--	public
	["GetSelectLimit"] = function(self)
		return self._selections:GetMaxSize();
	end,
	["SetSelectLimit"] = function(self, num)
		local overflows = self._selections:SetMaxSize(num);
		if overflows then
			for _, overflow in ipairs(overflows) do
				self:_dispatch(Selector_OnDeselected, overflow);
			end
		end
		self:Refresh();
	end,
	["IsDeselectable"] = function(self)
		return self._deselectable;
	end,
	["SetDeselectable"] = function(self, flag)
		self._deselectable = flag;
	end,
	["HasSelection"] = function(self, data)
		return self._selections:Exists(data);
	end,
	["RemoveSelection"] = function(self, data, motion)
		self._selections:Remove(data);
		self:_dispatch(Selector_OnDeselected, data, motion);
		self:Refresh();
	end,
	["AddSelection"] = function(self, data, motion)
		local overflows = self._selections:Add(data);
		if overflows then
			for _, overflow in ipairs(overflows) do
				self:_dispatch(Selector_OnDeselected, overflow);
			end
		end
		self:_dispatch(Selector_OnSelected, data, motion);
		self:Refresh();
	end,
	["GetSelection"] = function(self, index)
		return self._selections:Get(index);
	end,
	["ClearSelection"] = function(self)
		self._selections:Clear();
		for data, shine in pairs(self._shines) do
			returnShine(shine);
		end
		wipe(self._shines);
	end,
	["SetTemplate"] = function(self, name)
		if name == self._template then return; end
		self._template = name or DEFAULT_TEMPLATE;
		for index, e in ipairs(self._elements) do
			e:ClearAllPoints();
			e:Hide();
		end
		wipe(self._elements);
		for index, e in ipairs(self._pool) do
			e:ClearAllPoints();
			e:Hide();
		end
		wipe(self._pool);
		self:_rebuildElements(self._row, self._col);
		self:Refresh();
	end,
	["SetPreferedSize"] = function(self, row, col)
		local w = col * self.elementWidth + (col + 1) * self.elementGap + 0.1;
		local h = row * self.elementHeight + (row + 1) * self.elementGap + 0.1;
		self:SetSize(w, h);
		self:_updateViewport(row, col);
		self:Refresh();
		return w, h;
	end,
	["Refresh"] = function(self, massIndex, dataIndex)	--	updates elements & manipulates shines
		if self._dataAltered or not self._row or not self._col then
			self:_updateViewport();
		end
		if not massIndex or not dataIndex then
			massIndex, dataIndex = self:_getIndexByOffset(self.bar:GetValue());
		end
		if not massIndex or not dataIndex then return; end
		
		local col, row = self._col, self._row;
		local mass = self:_getMass(massIndex);
		local data = mass:Get(dataIndex);

		--	update elements
		wipe(self._elementDataMap);
		local skipped, notEmpty = 0;
		for i, e in ipairs(self._elements) do
			if skipped > 0 then			--	skip to newline (higher priority)
				skipped = skipped - 1;
				e:Hide();
			elseif data then			--	there is data, visualize it
				notEmpty = true;
				self:Visualize(e, data);
				self._elementDataMap[e] = data;
				--	selected shine
				if self:HasSelection(data) then
					self._shines[data] = applyShine(e, self._shines[data], self.selectedColor);
					self._shines[data].valid = true;
				end
				--	update mouseover
				pcall(e:IsMouseOver() and e:GetScript("OnEnter") or e:GetScript("OnLeave"), e, false);
				data = mass:Next(data);
				--	if is last data
				if not data and massIndex < #self._massList then
					massIndex = massIndex + 1;
					mass = self:_getMass(massIndex);
					data = mass:First();
					local leftRow = i % row;
					leftRow = leftRow == 0 and leftRow or row - leftRow;
					skipped = skipped + leftRow;
				end
			else
				e:Hide();
			end
		end
		--	recycle invalid shine
		for data, shine in pairs(self._shines) do
			if shine.valid then
				shine.valid = nil;
			else
				self._shines[data] = nil;
				returnShine(shine);
			end
		end
		--	empty indicator
		setVisible(self.emptyText, not notEmpty);
	end,
	["MountMassData"] = function(self, name, list, nonTerminated)
		local mass = MassData:New();
		mass:Consume(name, list);
		tinsert(self._massList, mass);
		self._dataAltered = true;
		if not nonTerminated then
			self:Refresh();
		end
	end,
	["ScrollToData"] = function(self, data)
		if not self:IsShown() then
			self._delayScroll = data;
		else
			self:_scrollTo(self:_getIndexByData(data));
		end
	end,
	["SetScrollValue"] = function(self, value, orientation)
		if not orientation or orientation == "HORIZONTAL" then
			self:_setOffset(value);
		end
	end,
	["HorizontalSmoothScrollTo"] = function(self, offset, duration)
		--	break & re-accelerate
		duration = duration or SCROLLTIME;
		offsetNew = mathBound(offset, self.bar:GetMinMaxValues());
		self._smoothVelocity = (offsetNew - self.bar:GetValue()) / duration;
		self._smoothTimer = duration;
	end,
	["HorizontalSmoothDeltaScroll"] = function(self, delta, duration)	--	up/left scroll is positive delta, down/right scroll is negative.
		duration = duration or SCROLLTIME;
		if not delta then
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		elseif not self._smoothTimer then
			--	initiating
			self._smoothVelocity = - delta * self.scrollStep / duration;
			self._smoothTimer = duration;
		elseif delta > 0 and 0 > self._smoothVelocity then
			--	accelerate backward
			self._smoothVelocity = self._smoothVelocity - delta * self.scrollStep / self._smoothTimer;
		elseif 0 > delta and self._smoothVelocity > 0 then
			--	accelerate forward
			self._smoothVelocity = self._smoothVelocity - delta * self.scrollStep / self._smoothTimer;
		else
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end,
	["Visualize"] = function(self, element, data)
		element.icon:SetTexture(data);
	end,
};

function MessFauxMassSelector_OnMouseWheel(self, delta)
	self:HorizontalSmoothDeltaScroll(delta);
end

function MessFauxMassSelector_OnSizeChanged(self, width, height)
	if not width or not height then return; end
	self.content:SetSize(width, height);
	local row = mathFloor((height - self.elementGap)/ (self.elementHeight + self.elementGap));
	local col = mathFloor((width - self.elementGap)/ (self.elementWidth + self.elementGap));
	self:_updateViewport(row, col);
	self:Refresh();
end

function MessFauxMassSelector_OnShow(self)
	if self._delayScroll then
		self:ScrollToData(self._delayScroll);
	end
	self._delayScroll = nil;
end

function MessFauxMassSelector_OnUpdate(self, elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	--	smooth scroll
	if self._smoothTimer then
		local value = self.bar:GetValue();

		value = value + mathMin(elapsed, self._smoothTimer) * self._smoothVelocity;
		self:_setOffset(value);

		self._smoothTimer = self._smoothTimer - elapsed;
		if 0 >= self._smoothTimer then
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end
end

function MessFauxMassSelector_OnLoad(self)
	MessEventDispatcher_OnLoad(self);
	inherits(self, MessFauxMassSelectorMeta);

	self._pool = {};
	self._selections = QueueTemplate:New();
	self._shines = {};
	self._elements = {};
	self._elementDataMap = {};
	self._massList = {};

	self.LBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.RBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.BBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.TBorder:SetVertexColor(unpack(COLOR_BORDER));

	self.bar:SetScrollParent(self);
	self.bar:SetOrientation("HORIZONTAL");
	self.bar:SetFadeAlpha(1, 0.5);
	self.bar:SetBorderColor(COLOR_BORDER);
end

