local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local COLOR_TEXT = {0.8, 0.8, 0.8};
local COLOR_FG = {1, 0.5, 0};
local COLOR_FG_FOCUS = {0.196, 0.196, 0.196};
local COLOR_FG_DISABLE = {0.4, 0.4, 0.4};
local COLOR_BG = {0.196, 0.196, 0.196};
local COLOR_BG_FOCUS = {1, 0.5, 0};
local COLOR_BG_DISABLE = {0.14, 0.14, 0.14};

local Item_OnFocusChanged = "Item_OnFocusChanged";

local dummy = function() return false; end

local function setVisible(region, visible)
	if visible then
		region:Show();
	else
		region:Hide();
	end
end

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

----------------------------------------------------------------	ScrollItemTemplate
local ScrollItemMeta = {
	["isFocused"] = nil,
	["normalBGColor"] = COLOR_BG,
	["focusBGColor"] = COLOR_BG_FOCUS,
	["disableBGColor"] = COLOR_BG_DISABLE,
	["normalFGColor"] = COLOR_FG,
	["focusFGColor"] = COLOR_FG_FOCUS,
	["disableFGColor"] = COLOR_FG_DISABLE,
	--	functions
	["Init"] = dummy,
	["HasFocus"] = function(self)
		return self.isFocused;
	end,
	["SetFocus"] = function(self, focus)
		if self.isFocused == focus then
			--	nothing
		else
			self.isFocused = focus;
			self:_dispatch(Item_OnFocusChanged, focus);
		end
		self:Update();
	end,
	["Update"] = function(self)
		if not self:IsEnabled() then
			self.bg:SetVertexColor(unpack(self.disableBGColor));
		elseif not self:HasFocus() then
			self.bg:SetVertexColor(unpack(self.normalBGColor));
		else
			self.bg:SetVertexColor(unpack(self.focusBGColor));
		end
	end,
};

function ScrollItemTemplate_OnLoad(self)
	inherits(self, ScrollItemMeta);
	MessEventDispatcher_OnLoad(self);
end

function ScrollItemTemplate_OnClick(self, ...)
	self:SetFocus(true);
end

function ScrollItemTemplate_OnEnter(self, motion)
	setVisible(self.highlight, self:IsMouseOver());
end

function ScrollItemTemplate_OnLeave(self, motion)
	setVisible(self.highlight, self:IsMouseOver());
end

function ScrollItemTemplate_OnEnable(self)
	self:Update();
end

function ScrollItemTemplate_OnDisable(self)
	self:Update();
end

----------------------------------------------------------------	ScrollSeparatorTemplate
local ScrollSeparatorMeta = {
	["normalFGColor"] = COLOR_TEXT,
	["Init"] = function(self, text)
		self.title:SetText(text);
	end,
	["HasFocus"] = dummy,
	["SetFocus"] = dummy,
	["ReleaseFocus"] = dummy,
	["Update"] = dummy,
};

function ScrollSeparator_OnLoad(self)
	inherits(self, ScrollSeparatorMeta);
	self:Update();
end

----------------------------------------------------------------	ScrollIconSelectorTemplate
local ScrollIconSelectorMeta = {
	--	overrides
	["Init"] = function(self, title, texture, func)
		self.title:SetText(title);
		self.icon:SetTexture(texture or "Interface\\Icons\\INV_Misc_QuestionMark");
		self.callbackFunc = func;
	end,
	["SetFocus"] = function(self, focus)
		if self.isFocused == focus then
			--	nothing
		else
			self.isFocused = focus;
			self:_dispatch(Item_OnFocusChanged, focus);
			if focus then
				self.iconSelector:BeginEditing(
					self.icon:GetTexture(), function(texture, motion)
						texture = texture or "Interface\\Icons\\INV_Misc_QuestionMark";
						self.icon:SetTexture(texture);
						pcall(self.callbackFunc, texture);
						self:SetFocus(false);
					end, 
					function()
						self:SetFocus(false);
					end, ANCHOR_RELATIVE, self.icon, 10, 5);
			else
				self.iconSelector:CancelEditing();
			end
		end
		self:Update();
	end,
	["Update"] = function(self)
		if not self:IsEnabled() then
			self.title:SetTextColor(unpack(self.disableFGColor));
			self.bg:SetVertexColor(unpack(self.disableBGColor));
		elseif self:HasFocus() then
			self.title:SetTextColor(unpack(self.focusFGColor));
			self.bg:SetVertexColor(unpack(self.focusBGColor));
		else
			self.title:SetTextColor(unpack(self.normalFGColor));
			self.bg:SetVertexColor(unpack(self.normalBGColor));
		end
	end,
	["SetIcon"] = function(self, texture)
		self.icon:SetTexture(texture or "Interface\\Icons\\INV_Misc_QuestionMark");
	end,
	["GetIcon"] = function(self)
		return self.icon:GetTexture();
	end,
};

function ScrollIconSelector_OnLoad(self)
	ScrollItemTemplate_OnLoad(self);
	inherits(self, ScrollIconSelectorMeta);
	self.iconSelector:SetPreferedSize(4, 6);
	self:Update();
end

----------------------------------------------------------------	ScrollTextfieldInputTemplate
local ScrollTextfieldInputMeta = {
	--	events
	["TextEdit_Modified"] = function(self, textfield, success)
		self:SetFocus(false);
		if success then
			pcall(self.callbackFunc, textfield:GetText());
		end
	end,
	["TextEdit_FocusChanged"] = function(self, textfield, focus)
		self:SetFocus(focus);
	end,
	--	overrides
	["Init"] = function(self, title, text, func)
		self.title:SetText(title);
		self:SetText(text);
		self.callbackFunc = func;
	end,
	["SetFocus"] = function(self, focus)
		if self.isFocused == focus then
			--	nothing
		else
			self.isFocused = focus;
			self:_dispatch(Item_OnFocusChanged, focus);
			if focus then
				self.textfield:SetFocus();
				self.textfield:HighlightText();
			else
				self.textfield:CancelEditing();
				self.textfield:HighlightText(1, 1);
			end
		end
		self:Update();
	end,
	["Update"] = function(self)
		if not self:IsEnabled() then
			self.title:SetTextColor(unpack(self.disableFGColor));
			self.bg:SetVertexColor(unpack(self.disableBGColor));
		elseif not self:HasFocus() then
			self.title:SetTextColor(unpack(self.normalFGColor));
			self.bg:SetVertexColor(unpack(self.normalBGColor));
		else
			self.title:SetTextColor(unpack(self.focusFGColor));
			self.bg:SetVertexColor(unpack(self.focusBGColor));
		end
	end,
	["SetText"] = function(self, text)
		self.textfield:BeginEditing(text or "");
	end,
	["GetText"] = function(self)
		return self.textfield:GetText();
	end,
};

function ScrollTextfieldInput_OnLoad(self)
	ScrollItemTemplate_OnLoad(self);
	inherits(self, ScrollTextfieldInputMeta);
	self:Update();
	
	self.textfield:RegisterCallback(self, "TextEdit_Modified");
	self.textfield:RegisterCallback(self, "TextEdit_FocusChanged");
end

----------------------------------------------------------------	ScrollTogglerTemplate
local ScrollTogglerMeta = {
	["Init"] = function(self, title, checked, func)
		self.title:SetText(title);
		self.callbackFunc = func;
		
		self:Update();
	end,
	["SetFocus"] = function(self, focus)
		if self.isFocused == focus then
			--	nothing
		else
			self.isFocused = focus;
			self:_dispatch(Item_OnFocusChanged, focus);
			pcall(self.callbackFunc, self:GetChecked());
		end
		self:Update();
	end,
	["Update"] = function(self)
		if not self:IsEnabled() then
			self.title:SetTextColor(unpack(self.disableFGColor));
			self.bg:SetVertexColor(unpack(self.disableBGColor));
			self.on:SetVertexColor(unpack(self.disableFGColor));
			self.off:SetVertexColor(unpack(disableFGColor));
		elseif not self:HasFocus() then
			self.title:SetTextColor(unpack(self.normalFGColor));
			self.bg:SetVertexColor(unpack(self.normalBGColor));
			self.on:SetVertexColor(unpack(self.normalFGColor));
			self.off:SetVertexColor(unpack(self.normalFGColor));
		else
			self.title:SetTextColor(unpack(self.focusFGColor));
			self.bg:SetVertexColor(unpack(self.focusBGColor));
			self.on:SetVertexColor(unpack(self.focusFGColor));
			self.off:SetVertexColor(unpack(self.focusFGColor));
		end
		local isChecked = self:GetChecked();
		setVisible(self.on, isChecked);
		setVisible(self.off, not isChecked);
	end,
	["SetTextures"] = function(self, onTexture, offTexture)
		self.on:SetTexture(onTexture);
		self.off:SetTexture(offTexture);
		self:Update();
	end,
};

function ScrollToggler_OnLoad(self)
	ScrollItemTemplate_OnLoad(self);
	inherits(self, ScrollTogglerMeta);
	
	hooksecurefunc(self, "SetChecked", function(self, checked)
		self:Update();
	end);
	
	self:Update();
end

----------------------------------------------------------------	ScrollColorPicker_OnLoad
local ScrollColorPickerMeta = {
	["_colorR"] = 1,
	["_colorG"] = 1,
	["_colorB"] = 1,
	["_colorA"] = 1,
	--	overrides
	["Init"] = function(self, color, title, func)
		self:SetColor(color);
		self.title:SetText(title);
		self.callbackFunc = func;
	end,
	["SetFocus"] = function(self, focus)
		if self.isFocused == focus then
			--	nothing
		else
			self.isFocused = focus;
			self:_dispatch(Item_OnFocusChanged, focus);
			if focus then
				--[[self.picker.previousValues = {self._colorR, self._colorG, self._colorB};
				self.picker.opacity = 1 - self._colorA;
				self.picker:SetColorRGB(self._colorR, self._colorG, self._colorB);
				--ColorPickerFrame.func, ColorPickerFrame.opacityFunc, ColorPickerFrame.cancelFunc ]]
				self.picker:Show();
			else
				self.picker:Hide();
			end
		end
		self:Update();
	end,
	["Update"] = function(self)
		if not self:IsEnabled() then
			self.title:SetTextColor(unpack(self.disableFGColor));
			self.bg:SetVertexColor(unpack(self.disableBGColor));
		elseif self:HasFocus() then
			self.title:SetTextColor(unpack(self.focusFGColor));
			self.bg:SetVertexColor(unpack(self.focusBGColor));
		else
			self.title:SetTextColor(unpack(self.normalFGColor));
			self.bg:SetVertexColor(unpack(self.normalBGColor));
		end
		self.color:SetVertexColor(self._colorR, self._colorG, self._colorB, self._colorA);
	end,
	["SetColor"] = function(self, color)
		if color.r and color.g and color.b then
			self._colorR, self._colorG, self._colorB, self._colorA = color.r, color.g, color.b, color.a;
		else
			self._colorR, self._colorG, self._colorB, self._colorA = unpack(color);
		end
		self._colorA = self._colorA or 1;
		self:Update();
	end,
	["GetColor"] = function(self)
		return self._colorR, self._colorG, self._colorB, self._colorA;
	end,
};

function ScrollColorPicker_OnLoad(self)
	ScrollItemTemplate_OnLoad(self);
	inherits(self, ScrollColorPickerMeta);
	
	self.picker.hasOpacity = true;
	
	self:Update();
end
