local mathMax = math.max;
local mathMin = math.min;
local mathFloor = math.floor;
local mathCeil = math.ceil;

local THRESHOLD = 0.01;
local SCROLLTIME = 0.3;
local LINEHEIHGT = 14;
local NEWLINE = "\n";

local COLOR_HIGHLIGHT = {1, 1, 1 , 0.2};
local COLOR_BORDER = {0.5, 0.5, 0.5, 1};
local COLOR_BG = {0.196, 0.196, 0.196};
local COLOR_FG = {1, 1, 1};
local COLOR_FG_DISABLED = {0.5, 0.5, 0.5};
local COLOR_INDICATOR = {0.25, 0.25, 0.25};

local TextEdit_Modified = "TextEdit_Modified";
local TextEdit_FocusChanged = "TextEdit_FocusChanged";

local function inherits(self, meta)
	for key, value in pairs(meta) do
		self[key] = value;
	end
end

----------------------------------------------------------------	MessHighlightTemplate
function MessEditHighlight_OnLoad(self)
	self.texture:SetVertexColor(unpack(COLOR_HIGHLIGHT));
end

function MessEditHighlight_OnEnter(self, motion)
	self.texture:Show();
	self.anime:Play();
end

function MessEditHighlight_OnLeave(self, motion)
	self.texture:Hide();
	self.anime:Stop();
end

function MessEditHighlight_OnMouseUp(self, ...)
	self:GetParent():SetFocus();
end

----------------------------------------------------------------	MessTextField
local MessTextFieldMeta = {
	["_ineditable"] = nil,
	["cancelWhileRelease"] = false,
	["_revertOnChar"] = function(self, text)
		local pos = self:GetCursorPosition();
		self:SetText(self.backup);
		self:SetCursorPosition(pos);
	end,
	["_revertOnTextChanged"] = function(self, userInput)
		if not userInput then return; end
		local pos = self:GetCursorPosition();
		self:SetText(self.backup);
		self:SetCursorPosition(pos);
	end,
	["BeginEditing"] = function(self, text)
		text = text or "";
		self.backup = text;
		self:SetText(text);
	end,
	["AcceptEditing"] = function(self)
		self.confirm = true;
		self:ClearFocus();
	end,
	["CancelEditing"] = function(self)
		self.cancel = true;
		self:ClearFocus();
	end,
	["IsEditable"] = function(self)
		return not self._ineditable;
	end,
	["SetEditable"] = function(self, flag)
		self._ineditable = not flag;
		if flag then
			self:SetTextColor(unpack(COLOR_FG));
			self:SetScript("OnChar", nil);
			self:SetScript("OnTextChanged", nil);
		else
			self:SetTextColor(unpack(COLOR_FG_DISABLED));
			self:SetScript("OnChar", self._revertOnChar);
			self:SetScript("OnTextChanged", self._revertOnTextChanged);
		end
	end,
};

function MessTextField_OnLoad(self)
	inherits(self, MessTextFieldMeta);
	MessEventDispatcher_OnLoad(self);
	self:SetTextColor(unpack(COLOR_FG));
	if self.bg then
		self.bg:SetVertexColor(unpack(COLOR_BG));
	end
	if self.underBorder then
		self.underBorder:SetVertexColor(unpack(COLOR_BORDER));
	end
end

function MessTextField_OnEscapePressed(self)
	self:CancelEditing();
end

function MessTextField_OnEnterPressed(self)
	self:AcceptEditing();
end

function MessTextField_OnEditFocusGained(self)
	self.cancel = nil;
	self.confirm = nil;
	if self.highlight then
		self.highlight:EnableMouse(false);
	end
	if self.bg then
		self.bg:Show();
	end
	self:_dispatch(TextEdit_FocusChanged, true);
end

function MessTextField_OnEditFocusLost(self)
	if self.cancel then
		self:SetText(self.backup or "");
		self:_dispatch(TextEdit_Modified);
	elseif self.confirm then
		self:_dispatch(TextEdit_Modified, true);
	else
		if self.cancelWhileRelease then
			self:SetText(self.backup or "");
		end
		self:_dispatch(TextEdit_Modified, not self.cancelWhileRelease);
	end
	if self.bg then
		self.bg:Hide();
	end
	if self.highlight then
		self.highlight:EnableMouse(true);
	end
	self:_dispatch(TextEdit_FocusChanged);
end

----------------------------------------------------------------	MessTextScrollFrameTemplate
local MessTextScrollerMeta = {
	--	attributes
	["scrollStep"] = LINEHEIHGT,
	["_smoothVelocity"] = 0,
	["_smoothTimer"] = nil,
	--	methods
	["SetFont"] = function(self, fileName, size, flag)
		self.textarea:SetFont(fileName, size, flag);
		self.scrollStep = size + self.textarea:GetSpacing();
	end,
	["SetFontObject"] = function(self, obj)
		self.textarea:SetFontObject(obj);
		local _, lineHeight = self.textarea:GetFont();
		self.scrollStep = lineHeight + self.textarea:GetSpacing();
	end,
	["SetText"] = function(self, text)
		self:ClearFocus();
		self.textarea:BeginEditing(text);
		self:VerticalSmoothScrollTo(0);
	end,
	["GetText"] = function(self)
		return self.textarea:GetText();
	end,
	["SetMaxLetters"] = function(self, limit)
		self.textarea:SetMaxLetters(limit);
	end,
	["GetMaxLetters"] = function(self)
		return self.textarea:GetMaxLetters();
	end,
	["SetFocus"] = function(self)
		self.textarea:SetFocus();
	end,
	["ClearFocus"] = function(self)
		self.textarea:ClearFocus();
	end,
	["IsEditable"] = function(self)
		return self.textarea:IsEditable();
	end,
	["SetEditable"] = function(self, flag)
		self.textarea:SetEditable(flag);
	end,
	["RegisterCallback"] = function(self, listener, event)
		self.textarea:RegisterCallback(listener, event);
	end,
	["UnregisterCallback"] = function(self, listener, event)
		self.textarea:UnregisterCallback(listener, event);
	end,
	["UnregisterAllCallbacks"] = function(self, listener)
		self.textarea:UnregisterAllCallbacks(listener);
	end,
	["VerticalSmoothDeltaScroll"] = function(self, delta, duration)
		duration = duration or SCROLLTIME;
		if not self._smoothTimer then
			self._smoothVelocity = -delta * self.scrollStep / duration;
			self._smoothTimer = duration;
		elseif delta > 0 and 0 > self._smoothVelocity then
			--	accelerate
			self._smoothVelocity = self._smoothVelocity - self.scrollStep / self._smoothTimer;
		elseif 0 > delta and self._smoothVelocity > 0 then
			--	accelerate
			self._smoothVelocity = self._smoothVelocity + self.scrollStep / self._smoothTimer;
		else
			--	break & stop
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end,
	["VerticalSmoothScrollTo"] = function(self, offset, duration)
		--	break & re-accelerate
		duration = duration or SCROLLTIME;
		local value = self:GetVerticalScroll();
		local range = self:GetVerticalScrollRange();
		offset = mathMax(0, mathMin(range, offset));
		self._smoothVelocity = (offset - value) / duration;
		self._smoothTimer = duration;
	end,
	["SetScrollValue"] = function(self, value, orientation)	-- adaptor for MessScrollBarTemplate
		if not orientation or orientation == "VERTICAL" then
			self:VerticalSmoothScrollTo(value);
		end
	end,
};

function MessTextScroller_OnLoad(self)
	inherits(self, MessTextScrollerMeta);

	self.LBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.RBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.BBorder:SetVertexColor(unpack(COLOR_BORDER));
	self.TBorder:SetVertexColor(unpack(COLOR_BORDER));
	
	self.bar:SetScrollParent(self);
	self.bar:SetFadeAlpha(1, 0.5);
end

function MessTextScroller_OnSizeChanged(self, w, h)	--todo
	self.bar:SetViewportSize(w, h);
	self.textarea:SetWidth(w);
end

function MessTextScroller_OnScrollRangeChanged(self, xrange, yrange)
	self.bar:SetMinMaxValues(0, yrange);
end

function MessTextScroller_OnVerticalScroll(self, offset)
	self.bar:SetValue(offset);
end

function MessTextScroller_OnMouseWheel(self, delta)
	self:VerticalSmoothDeltaScroll(delta);
end

function MessTextScroller_OnShow(self)
	self._elapsed = 0;
end

function MessTextScroller_OnUpdate(self, elapsed)
	self._elapsed = self._elapsed + elapsed;
	if THRESHOLD > self._elapsed then return end
	elapsed = mathFloor(self._elapsed / THRESHOLD) * THRESHOLD;
	self._elapsed = self._elapsed - elapsed;
	-- smooth scroll
	if self._smoothTimer then
		local value = self:GetVerticalScroll();
		local range = self:GetVerticalScrollRange();

		value = value + mathMin(elapsed, self._smoothTimer) * self._smoothVelocity;
		value = mathMax(0, mathMin(range, value));
		self:SetVerticalScroll(value);

		self._smoothTimer = self._smoothTimer - elapsed;
		if 0 >= self._smoothTimer then
			self._smoothVelocity = 0;
			self._smoothTimer = nil;
		end
	end
end

----------------------------------------------------------------	Textarea (insider of MessTextScrollFrameTemplate)
function MessTextarea_OnLoad(self)
	inherits(self, MessTextFieldMeta);
	MessEventDispatcher_OnLoad(self);
	self:SetTextColor(unpack(COLOR_FG));
	self.indicator:SetVertexColor(unpack(COLOR_INDICATOR));
end

function MessTextarea_OnEscapePressed(self)
	self:CancelEditing();
end

function MessTextarea_OnEnterPressed(self)
	self:Insert(NEWLINE);
end

function MessTextarea_OnEditFocusGained(self)
	self.cancel = nil;
	self.confirm = nil;
	self.indicator:Show();
	
	local parent = self:GetParent();
	parent.highlight:EnableMouse(false);
	
	self:_dispatch(TextEdit_FocusChanged, true);
end

function MessTextarea_OnEditFocusLost(self)
	if self.cancel then
		self:SetText(self.backup or "");
		self:_dispatch(TextEdit_Modified);
	elseif self.confirm then
		self:_dispatch(TextEdit_Modified, true);
	else
		if self.cancelWhileRelease then
			self:SetText(self.backup or "");
		end
		self:_dispatch(TextEdit_Modified, not self.cancelWhileRelease);
	end
	self.indicator:Hide();
	
	local parent = self:GetParent();
	parent.highlight:EnableMouse(true);
	
	self:_dispatch(TextEdit_FocusChanged);
end

function MessTextarea_OnCursorChanged(self, x, y, w, h)
	local parent = self:GetParent();
	y = -y;
	
	if self:HasFocus() then	--	moved by user action
		--	scroll to cursor
		local height = parent:GetHeight();
		local range = parent:GetVerticalScrollRange();
		local value = parent:GetVerticalScroll();
		if mathFloor(range) > 0 and mathFloor(height) > 0 then
			if y < value then	
				--	upward
				value = mathFloor(y / parent.scrollStep) * parent.scrollStep;
				parent:VerticalSmoothScrollTo(mathMax(0, mathMin(value, range)));
			elseif y + h > value + height then
				--	downward
				value = mathCeil((y + h) / parent.scrollStep) * parent.scrollStep - height;
				parent:VerticalSmoothScrollTo(mathMax(0, mathMin(value, range)));
			end
		end
		--	line indicator (exprimental)
		local pos = self:GetCursorPosition();
		local text = self:GetText();
		local lineList = {strsplit("\n", text)};
		
		local foregoingLength, contextLength = 0, 0;
		local count = 0;
		for index, line in ipairs(lineList) do
			local length = strlen(line);
			
			if pos > length then
				count = count + length + 1;
				pos = pos - length - 1;
			else
				foregoingLength = count - 1;
				contextLength = count + length;
				break;
			end
		end
		
		self.measurement1:SetWidth(self:GetWidth());
		if foregoingLength == -1 then
			self.measurement1:SetText("");
		else
			self.measurement1:SetText(strsub(text, 0, foregoingLength));
		end
		if strlen(self.measurement1:GetText()) == 0 then
			self.indicator:SetPoint("TOP");
		else
			self.indicator:SetPoint("TOP", self.measurement1, "BOTTOM");
		end
		self.measurement2:SetWidth(self:GetWidth());
		self.measurement2:SetText(strsub(text, 0, contextLength));
		self.indicator:SetPoint("BOTTOM", self.measurement2, "BOTTOM");
	end
end
