local _name, _addon = ...;

local BASE_TYPE = _addon.BASE_TYPE;
local EXTRA_TYPE = _addon.EXTRA_TYPE;

local Dewdrop = LibStub("Dewdrop-2.0");
local LDB = LibStub("LibDataBroker-1.1");

local ICON_DEFAULT = "Interface\\Icons\\INV_Misc_QuestionMark";

---------------------------------------------------------------------------------	caches
local CACHE = {
	root = nil,
	frame = nil,
	showFolderIcon = nil,
	persisted = nil,
	hideMissing = nil,
	hideCooling = nil,
	entryHeight = 16,
	textHeight = 14,
	separatorHeight = 7,
	colorDisabled = GRAY_FONT_COLOR,
	colorMissing = DIM_RED_FONT_COLOR,
	colorCooldown = LIGHTYELLOW_FONT_COLOR,
	colorEquipped = GREEN_FONT_COLOR,
	colorEquippable = ORANGE_FONT_COLOR,
	colorNormal = HIGHLIGHT_FONT_COLOR,
}
---------------------------------------------------------------------------------
local dummy = function() return; end

local refresher = function(lv)
	Dewdrop:Refresh(lv);
end

local function rgbUnpack(color)
	return color.r, color.g, color.b;
end

---------------------------------------------------------------------------------
local MenuItemPolish = {
	[BASE_TYPE["COMPANION"]] = function(self, entry)
		local spellID = entry.data;
		self.isMissing = false; --not IsSpellKnown(spellID); -- companion spell is alyways unknown
		local start, duration, enable = GetSpellCooldown(spellID);
		local TTL = start and (start + duration - GetTime()) or -1;
		self.isCooling = TTL > 0;
		self.disabled = self.isMissing or self.isCooling;
		self.secure = {
			type = "macro",
			macrotext = entry.cmd,
		};
		return self;
	end,
	[BASE_TYPE["BATTLEPET"]] = function(self, entry)
		local petGUID = entry.data;
		self.isMissing = not C_PetJournal.GetPetInfoByPetID(petGUID);
		self.disabled = self.isMissing;
		self.arg2 = petGUID;
		self.func = function(lv, id)
			C_PetJournal.SummonPetByGUID(id);
			Dewdrop:Refresh(lv);
		end
		return self;
	end,
	[BASE_TYPE["SPELL"]] = function(self, entry)
		local spellID = entry.data;
		self.isMissing = not IsSpellKnown(spellID);
		if not self.isMissing then
			local start, duration, enable = GetSpellCooldown(spellID);
			local TTL = start and (start + duration - GetTime()) or -1;
			self.isCooling = TTL > 0;
		end
		self.disabled = self.isMissing or self.isCooling;
		self.secure = {
			type = "macro",
			macrotext = entry.cmd,
		};
		return self;
	end,
	[BASE_TYPE["ITEM"]] = function(self, entry)
		local itemLink = entry.data;
		local itemID = select(3, strfind(itemLink, "^|%x+|Hitem:(%-?%d+):(%d+):(%d+):(%d+):(%d+):(%d+):(%-?%d+):(%-?%d+)"));
		self.isMissing = true;
		if PlayerHasToy(itemID) then
			self.isMissing = false;
		else
			self.isMissing = GetItemCount(itemLink) <= 0;
		end
		if not self.isMissing then
			local start, duration, enable = GetItemCooldown(itemID);
			local TTL = start and (start + duration - GetTime()) or -1;
			self.isCooling = TTL > 0;
		end
		self.disabled = self.isMissing or self.isCooling;
		self.isEquipped = IsEquippedItem(itemLink);
		self.isEquippable = IsEquippableItem(itemLink);
		self.closeWhenClicked = not (self.isEquippable and not self.isEquipped);
		self.secure = {
			type = "macro",
			macrotext = entry.cmd,
		};
		return self;
	end,
	[BASE_TYPE["MACRO"]] = function(self, entry)
		local macroID = entry.data;
		self.isMissing = self.char and self.char ~= GetUnitName("player", true);
		if not self.isMissing then
			local t = GetTime();
			local _, _, spellID = GetMacroSpell(macroID);
			local s, d = GetSpellCooldown(spellID);
			local TTL_spell = s and (s + d - t) or -1;
			local _, itemID = GetMacroItem(macroID);
			s, d = GetItemCooldown(itemID);
			local TTL_item = s and (s + d - t) or -1;
			self.isCooling = TTL_spell > 0 or TTL_item > 0;
		end
		self.disabled = self.isMissing or self.isCooling;
		self.secure = {
			type = "macro",
			macrotext = entry.cmd,
		};
		return self;
	end,
	[BASE_TYPE["EQUIPMENTSET"]] = function(self, entry)
		local setName = entry.data;
		if CanUseEquipmentSets() then
			local icon, _, isEquipped, numItems, numEquipped, _, numMissing, numIgnored = GetEquipmentSetInfoByName(setName);
			self.icon = icon or self.icon;
			self.isMissing = not icon or numMissing > 0;
			self.isEquipped = numEquipped == numItems;
			self.arg2 = setName;
			self.func = function(lv, name)
				UseEquipmentSet(name);
				Dewdrop:Refresh(lv);
			end
		else
			self.disabled = true;
		end
		self.isEquippable = true;
		return self;
	end,
	[EXTRA_TYPE["CATEGORY"]] = function(self, entry)
		local hasChildren = entry.subList and #entry.subList > 0;
		self.hasArrow = hasChildren;
		self.subMenu = hasChildren and {};
		return self;
	end,
	[EXTRA_TYPE["ADDONMACRO"]] = function(self, entry)
		self.secure = {
			type = "macro",
			macrotext = entry.cmd,
		};
		return self;
	end,
	[EXTRA_TYPE["LAUNCHER"]] = function(self, entry, frame)
		local brokerName = entry.data;
		local obj = LDB:GetDataObjectByName(brokerName);
		self.icon = obj.icon;
		self.isMissing = not obj;
		self.disabled = self.isMissing or not obj.OnClick;
		self.func = function()
			pcall(obj.OnClick, frame, entry.action);	-- dewdrop menu GetMouseButtonClicked() = LeftButton
		end;
		return self;
	end,
	[EXTRA_TYPE["SEPARATOR"]] = function(self)
		self.isSeparator = true;
		self.textHeight = CACHE.separatorHeight;
		self.disabled = true;
		return self;
	end,
};

---------------------------------------------------------------------------------
local function parseEntries(container, root, level)
	for _, entry in ipairs(root.subList) do
		if entry.isGarbage then
			--	skip
		elseif entry.outOfSync then
			--	skip
		else
			--	default
			local item = {
				iconHeight = CACHE.entryHeight,
				iconWidth = CACHE.entryHeight,
				textHeight = CACHE.textHeight,
				text = entry.name,
				icon = (entry.type ~= EXTRA_TYPE["CATEGORY"] or CACHE.showFolderIcon) and entry.texture,
				closeWhenClicked = not CACHE.persisted,
				tooltipFunc = dummy,
				arg1 = level,
				func = refresher,
			};
			--	polish
			if MenuItemPolish[entry.type] then
				item = MenuItemPolish[entry.type](item, entry, CACHE.frame);
			end
			--	coloring
			if item.disabled then
				item.textR = CACHE.colorDisabled.r;
				item.textG = CACHE.colorDisabled.g;
				item.textB = CACHE.colorDisabled.b;
			elseif item.isMissing then
				item.textR = CACHE.colorMissing.r;
				item.textG = CACHE.colorMissing.g;
				item.textB = CACHE.colorMissing.b;
			elseif item.isCooling then
				item.textR = CACHE.colorCooldown.r;
				item.textG = CACHE.colorCooldown.g;
				item.textB = CACHE.colorCooldown.b;
			elseif item.isEquipped then
				item.textR = CACHE.colorEquipped.r;
				item.textG = CACHE.colorEquipped.g;
				item.textB = CACHE.colorEquipped.b;
			elseif item.isEquippable then
				item.textR = CACHE.colorEquippable.r;
				item.textG = CACHE.colorEquippable.g;
				item.textB = CACHE.colorEquippable.b;
			else
				item.textR = CACHE.colorNormal.r;
				item.textG = CACHE.colorNormal.g;
				item.textB = CACHE.colorNormal.b;
			end
			--	visibility
			if CACHE.hideMissing and item.isMissing then
				--	skip
			elseif CACHE.hideCooling and item.isCooling then
				--	skip
			else
				tinsert(container, item);
				if item.subMenu then
					parseEntries(item.subMenu, entry, level + 1);
				end
			end
		end
	end
	local emptySlot = {};
	--	remove heading seprators
	for index = 1, #container, 1 do
		local item = container[index];
		if item.isSeparator then
			tinsert(emptySlot, 1, index);
		else
			break;
		end
	end
	--	remove tailing seprators
	for index = #container, 1, -1 do
		local item = container[index];
		if item.isSeparator then
			tinsert(emptySlot, 1, index);
		else
			break;
		end
	end
	--	do removal
	for _, index in ipairs(emptySlot) do
		tremove(container, index);
	end
	--	done
	return container;
end

---------------------------------------------------------------------------------
function _addon:Menu_Feed(relatedFrame, root, level)
	CACHE.root = root;
	CACHE.frame = relatedFrame;
	CACHE.showFolderIcon = root.showFolderIcon;
	CACHE.persisted = root.persisted;
	CACHE.hideMissing = root.hideMissing;
	CACHE.hideCooling = root.hideCooling;
	
	local menuInfo = {};
	parseEntries(menuInfo, root, level);
	Dewdrop:FeedTable(menuInfo);
end
