local _name, _addon = ...;
local BASE_TYPE = _addon.BASE_TYPE;
local EXTRA_TYPE = _addon.EXTRA_TYPE;

---------------------------------------------------------------------------------	macro
-- MAX_ACCOUNT_MACROS = 36
-- MAX_CHARACTER_MACROS = 18
local entryMacroIDMonitor = {};	--	entry : macroID

local function matchMacro(entry, macroID)
	local name, texture, body = GetMacroInfo(macroID);
	if entry.name == name and entry.texture == texture and entry.cmd == body then
		return macroID;
	end
end

local function findEditedMacroID(id, name, texture, body)
	--	fast try
	local n, t, b = GetMacroInfo(id);
	if n and n == name and t == texture and b == body then
		return id;
	end
	--	one by one
	for index = 1, MAX_ACCOUNT_MACROS do
		n, t, b = GetMacroInfo(index);
		if n and n == name and t == texture and b == body then
			return index;
		end
	end
end

local function hookMacroFuncs()
	--	build cache & checking
	local nameMap = {};		--	name : marcoID/macroID list (duplicated name list)
	for index = 1, MAX_ACCOUNT_MACROS do
		local name, texture, body = GetMacroInfo(index);
		if not name then
			--	no macro
		elseif nameMap[name] then
			print("warning : mutiple macros with identical name '"..name.."' detected. please rename for better competibility.");
			if type(nameMap[name]) == "table" then
				tinsert(nameMap[name], index);
			else
				local tmp = {};
				tinsert(tmp, nameMap[name]);
				nameMap[name] = tmp;
			end
		else
			nameMap[name] = index;
		end
	end
	--	valid current monitoring
	for entry, macroID in pairs(entryMacroIDMonitor) do
		if not nameMap[entry.name] then
			--	out of sync
			entry.data = nil;
			entry.outOfSync = true;
			entryMacroIDMonitor[entry] = nil;
		else
			local macroIDs = nameMap[entry.name];
			if type(macroIDs) == "table" then
				for _, macroID in ipairs(macroIDs) do
					entry.data = matchMacro(entry, macroID);
					if entry.data then
						break;
					end
				end
			else
				entry.data = matchMacro(entry, macroIDs);
			end
			entry.outOfSync = not entry.data;
		end
	end
	--	hooks
	--[[--	create new macro, which has no interference with any existed entry, ignore it.
	hooksecurefunc("CreateMacro", function(name, texture, body, charSpecific)
	end);]]
	hooksecurefunc("EditMacro", function(id, name, texture, body)
		--	use back-tracing, thus we dont taint 'EditMacro'
		local newID = findEditedMacroID(id, name, texture, body);
		for entry, macroID in pairs(entryMacroIDMonitor) do
			if macroID == newID then
				entry.name = name;
				entry.texture = texture;
				entry.cmd = body;
				entry.data = newID;
			end
		end
	end);
	hooksecurefunc("DeleteMacro", function(target)
		if type(target) == "number" then
			for entry, macroID in pairs(entryMacroIDMonitor) do
				if macroID == target then
					entry.data = nil;
					entry.outOfSync = true;
					entryMacroIDMonitor[entry] = nil;
				end
			end
		elseif type(target) == "string" then
			for entry, _ in pairs(entryMacroIDMonitor) do
				if entry.name == target then
					local macroID = matchMacro(entry, entry.data);
					if not macroID then
						entry.data = nil;
						entry.outOfSync = true;
						entryMacroIDMonitor[entry] = nil;
					end
				end
			end
		end
	end);
end

---------------------------------------------------------------------------------	equipment set
--	equipment set indexing by name, monitoring name chnages
local entryEquipmentSetNameMonitor = {};	-- entry : set name

local function hookEquipmentSetFuncs()
	--	hooks
	--[[
	--	equipmentset creation or content update
	--	for creation, which has no interference with any existed entry, ignore it.
	--	for content update, as long as they have identical name, do nothing.
	hooksecurefunc("SaveEquipmentSet", function(name, texture)
	end);]]
	hooksecurefunc("ModifyEquipmentSet", function(oldName, newName)
		for entry, name in pairs(entryEquipmentSetNameMonitor) do
			if name == oldName then
				entry.name = entry.name == oldName and newName;
				entry.data = newName;
				entryEquipmentSetNameMonitor[entry] = newName;
			end
		end
	end);
	hooksecurefunc("DeleteEquipmentSet", function(delName)
		for entry, name in pairs(entryEquipmentSetNameMonitor) do
			if name == delName then
				entry.data = nil;
				entry.outOfSync = true;
			end
		end
	end);
end

---------------------------------------------------------------------------------
-- only interested entry would be cached (account wide, current character)
local interestedCases = { 
	[BASE_TYPE["MACRO"]] = function(entry)
		if entry.char and entry.char ~= GetUnitName("player", true) then
			--	other char
		elseif not entry.data or entry.outOfSync then
			--	out of sync
		else
			entryMacroIDMonitor[entry] = entry.data;
		end
	end,
	[BASE_TYPE["EQUIPMENTSET"]] = function(entry)
		if entry.char and entry.char == GetUnitName("player", true) then
			--	other char
		elseif not entry.data or entry.outOfSync then
			--	out of sync
		else
			entryEquipmentSetNameMonitor[entry] = entry.name;
		end
	end,
};

local function buildCaches(root)
	if not root.subList then return; end
	for _, entry in ipairs(root.subList) do
		if interestedCases[entry.type] then
			interestedCases[entry.type](entry);
		end
		if entry.subList then
			buildCaches(entry);
		end
	end
end

---------------------------------------------------------------------------------	global
function _addon:Sync_HookMacroFuncs()
	hookMacroFuncs();
end

function _addon:Sync_HookEquipmentSetFunc()
	hookEquipmentSetFuncs();
end

function _addon:Sync_CacheEntry(entry)
	buildCaches(entry);
end

function _addon:Sync_DecacheEntry(entry)	--	actually, there's no necessarity to decache each removed entry, doh.
	entryMacroIDMonitor[entry] = nil;
	entryEquipmentSetNameMonitor[entry] = nil;
	if entry.subList then
		for _, subentry in ipairs(entry.subList) do
			self:Sync_DecacheEntry(subentry);
		end
	end
end

